@@ -1,5 +1,23 @@
 /*
  * very very simple makefile parser
+ *
+ * Copyright (c) 2003 SuSE Linux AG
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
+ *  
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program (see the file COPYING); if not, write to the
+ * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
+ *
  */
 
 #include <stdio.h>
@@ -248,6 +266,9 @@ void parse_makefile(const char *path)
 		if (! strcmp(node->name, "xdm"))
 			importance = 100;
 
+		if (! strcmp(node->name, "sshd"))
+			importance = 2000;
+
 		if (! strncmp(node->name, "early", 5))
 			importance = 8000;
 
@@ -388,7 +409,6 @@ void check_run_files(const char *action, const char *prev, const char *run)
 # define bootlog(arg...)
 #endif
 
-
 /*
  * pick up the next running task
  * return NULL if not found.
@@ -398,10 +418,12 @@ struct makenode *pickup_task(void)
 	struct makenode *node, *best = (struct makenode*)0;
 
 	for (node = tree_list; node; node = node->next) {
-		if ((! node->status) && (! node->num_deps) &&
-		    ((! best) || (node->importance > best->importance))) {
+		if (node->status != T_READY)
+			continue;
+		if (node->num_deps > 0)
+			continue;
+		if (!best || (node->importance > best->importance))
 			best = node;
-		}
 	}
 	if (best) {
 #if defined _XOPEN_SOURCE && (_XOPEN_SOURCE - 0) >= 600
