@@ -31,6 +31,7 @@
 #endif
 #include <stdio.h>
 #include <termios.h>
+#include <signal.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/select.h>
@@ -40,6 +41,8 @@
 #include <sys/un.h>
 #include <sys/sysinfo.h>
 #include <sys/stat.h>
+#include <sys/mman.h>
+#include <sys/signalfd.h>
 #ifdef CHECK_FORDEVPTS
 # include <sys/vfs.h>
 #endif
@@ -52,7 +55,25 @@
 #include "makeboot.h"
 #include "proc.h"
 
-#define timerdiff(n,l) (__extension__ ({ (((n).tv_sec-(l).tv_sec)*1000)+(((n).tv_usec-(l).tv_usec)/1000); }))
+#if !defined(__STDC_VERSION__) || (__STDC_VERSION__ < 199901L)
+# ifndef  inline
+#  define inline		__inline__
+# endif
+# ifndef  restrict
+#  define restrict		__restrict__
+# endif
+# ifndef  volatile
+#  define volatile		__volatile__
+# endif
+# ifndef  extension
+#  define extension		__extension__
+# endif
+#endif
+#ifndef  attribute
+# define attribute(attr)	__attribute__(attr)
+#endif
+
+#define timerdiff(n,l) (extension({ (((n).tv_sec-(l).tv_sec)*1000)+(((n).tv_usec-(l).tv_usec)/1000); }))
 
 typedef enum _boolean {false, true} boolean;
 extern char *optarg;
@@ -61,7 +82,7 @@ extern int optind;
 static long int numcpu = -1;
 static char *myname;
 static struct termios tio;
-static struct winsize wz;
+static volatile struct winsize wz;
 static struct {
   char env_row[128];
   char env_col[128];
@@ -75,8 +96,11 @@ static struct timeval now;
 static struct timeval lastlim;
 static char *run_mode = NULL;
 static struct makenode **nodevec;
+static sig_atomic_t signaled;
 
 #define PBUF_SIZE	8192
+#define PRUNNING	0x0001
+#define PFINISHED	0x0002
 struct prg {
   const char *name;
   const char *arg0;
@@ -84,25 +108,26 @@ struct prg {
   int fd;
   pid_t pid;
   struct timeval lastio;
-  size_t len;
   int splashadd;
+  int status;
+  int flags;
+  size_t len;
   char buf[PBUF_SIZE];
 };
 
 static struct prg *prgs;
 static int inpar, par;
-static int pidpipe[2];
 static double iorate = 800.0;
 
 #ifdef USE_PRELOAD
-static enum { Unknown, Preload, NoPreload } ispreload = Unknown;
+static volatile enum { Unknown, Preload, NoPreload } ispreload = Unknown;
 
-static void sighandler_nopreload(int x __attribute__((unused)))
+static void sighandler_nopreload(int x attribute((unused)))
 {
     ispreload = NoPreload;
 }
 
-static void sighandler_preload(int x __attribute__((unused)))
+static void sighandler_preload(int x attribute((unused)))
 {
     ispreload = Preload;
 }
@@ -160,10 +185,7 @@ void waitsplash()
   int status;
   if (!splashpid)
     return;
-  do {
-    waitpid(splashpid, &status, 0);
-  } while (errno == EINTR);
-  splashpid = 0;
+  TEMP_FAILURE_RETRY(waitpid(splashpid, &status, 0));
 }
 
 void closeall(void)
@@ -175,8 +197,6 @@ void closeall(void)
   for (s = 0; s < par; s++)
     if (prgs[s].fd)
       close(prgs[s].fd);
-  close(pidpipe[0]);
-  close(pidpipe[1]);
 }
 
 void callsplash(int n, const char *path, char *action)
@@ -186,6 +206,7 @@ void callsplash(int n, const char *path, char *action)
   char tbuf[256];
   pid_t pid;
   struct stat stb;
+  sigset_t nmask;
 
   if (n < 0 || splashpos < 0)
     return;
@@ -219,8 +240,26 @@ void callsplash(int n, const char *path, char *action)
       splashpid = pid;
       return;
     }
-  while (dup2(2, 1) < 0 && (errno == EINTR))
-    ;
+
+  (void)sigemptyset(&nmask);
+  (void)sigaddset(&nmask, SIGINT);
+  (void)sigaddset(&nmask, SIGHUP);
+  (void)sigaddset(&nmask, SIGQUIT);
+  (void)sigaddset(&nmask, SIGSEGV);
+  (void)sigaddset(&nmask, SIGTERM);
+  (void)sigaddset(&nmask, SIGCHLD);
+  (void)sigaddset(&nmask, SIGTTIN);
+  sigprocmask(SIG_UNBLOCK, &nmask, NULL);
+
+  (void)signal(SIGINT,  SIG_DFL);
+  (void)signal(SIGHUP,  SIG_DFL);
+  (void)signal(SIGQUIT, SIG_DFL);
+  (void)signal(SIGSEGV, SIG_DFL);
+  (void)signal(SIGTERM, SIG_DFL);
+  (void)signal(SIGCHLD, SIG_DFL);
+  (void)signal(SIGTTIN, SIG_DFL);
+
+  TEMP_FAILURE_RETRY(dup2(2, 1));
   closeall();
   execl("/sbin/splash", "splash", "-p", sbuf, "-t", tbuf, splashcfg, (char *)0);
   _exit(1);
@@ -236,6 +275,8 @@ void writebuf(struct prg *p)
       r = write(2, b, p->len);
       if (r < 0)
 	{
+	  if (errno == EINTR)
+	    continue;
 	  perror("write");
 	  r = p->len;
 	}
@@ -283,8 +324,8 @@ static int checksystem(const int par, const boolean start, const boolean limit)
     prcs_blked = 0;
 #endif
 
-  newpar  = (par*numcpu) - prcs_run + 1; 	/* +1 for startpar its self */
-  newpar -= (int)(((double)prcs_blked)*iorate);	/* I/O load reduction */
+  newpar  = (par*numcpu) - (prcs_run + prcs_blked);
+  newpar -= ((int)(((double)prcs_blked)*iorate))/1000;	/* I/O load reduction */
 
 #if DEBUG
   fprintf(stderr, "checksystem par=%d newpar=%d (prcs_run=%lu) %ld\n", par, newpar, prcs_run, time(0));
@@ -363,7 +404,7 @@ static int checkdevpts(void)
     }
 }
 #else
-static int __attribute__ ((unused)) checkdevpts(void)
+static int attribute((unused)) checkdevpts(void)
 {
   return 1;	/* /dev/pts is always mounted */
 }
@@ -372,6 +413,7 @@ static int __attribute__ ((unused)) checkdevpts(void)
 void run(struct prg *p)
 {
   const char * m = (char*)0;
+  sigset_t nmask;
 #ifdef USE_PRELOAD
   pid_t parent;
 #endif
@@ -398,38 +440,52 @@ void run(struct prg *p)
     {
       perror("fork");
       fprintf(stderr, "could not fork %s\n", p->name);
-      p->pid = 0;
+      p->pid = (pid_t)0;
       if (p->fd)
 	{
 	  close(p->fd);
 	  p->fd = 0;
 	}
+      p->status = 0;
+      p->flags = 0;
       return;
     }
   if (p->pid != 0)
-    return;
+    {
+      p->flags |= PRUNNING;
+      return;
+    }
+
+  (void)sigemptyset(&nmask);
+  (void)sigaddset(&nmask, SIGINT);
+  (void)sigaddset(&nmask, SIGHUP);
+  (void)sigaddset(&nmask, SIGQUIT);
+  (void)sigaddset(&nmask, SIGSEGV);
+  (void)sigaddset(&nmask, SIGTERM);
+  (void)sigaddset(&nmask, SIGCHLD);
+  (void)sigaddset(&nmask, SIGTTIN);
+  sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
   (void)signal(SIGINT,  SIG_DFL);
+  (void)signal(SIGHUP,  SIG_DFL);
   (void)signal(SIGQUIT, SIG_DFL);
   (void)signal(SIGSEGV, SIG_DFL);
   (void)signal(SIGTERM, SIG_DFL);
   (void)signal(SIGCHLD, SIG_DFL);
+  (void)signal(SIGTTIN, SIG_DFL);
 
   if (setpgid(0, 0))
     perror("setpgid");
 
   if (m && p->fd)
     {
-      while (close(1) < 0 && (errno == EINTR))
-	;
+      TEMP_FAILURE_RETRY(close(1));
       if (open(m, O_RDWR) != 1)
 	{
 	  perror(m);
 	  _exit(1);
 	}
-      while (dup2(1, 2) < 0 && (errno == EINTR))
-	;
-      tio.c_oflag &= ~OPOST;
+      TEMP_FAILURE_RETRY(dup2(1, 2));
       if (tcsetattr(1, TCSANOW, &tio))
 	perror("tcsetattr");
       if (wzok)
@@ -439,8 +495,7 @@ void run(struct prg *p)
     }
   else
     {
-      while (dup2(2, 1) < 0 && (errno == EINTR))
-	;
+      TEMP_FAILURE_RETRY(dup2(2, 1));
     }
 
   closeall();
@@ -525,14 +580,27 @@ int run_single(const char *prg, const char *arg0, int spl)
 
   if (pid == 0)
     {
+      sigset_t nmask;
+
+      (void)sigemptyset(&nmask);
+      (void)sigaddset(&nmask, SIGINT);
+      (void)sigaddset(&nmask, SIGHUP);
+      (void)sigaddset(&nmask, SIGQUIT);
+      (void)sigaddset(&nmask, SIGSEGV);
+      (void)sigaddset(&nmask, SIGTERM);
+      (void)sigaddset(&nmask, SIGCHLD);
+      (void)sigaddset(&nmask, SIGTTIN);
+      sigprocmask(SIG_UNBLOCK, &nmask, NULL);
+
       (void)signal(SIGINT,  SIG_DFL);
+      (void)signal(SIGHUP,  SIG_DFL);
       (void)signal(SIGQUIT, SIG_DFL);
       (void)signal(SIGSEGV, SIG_DFL);
       (void)signal(SIGTERM, SIG_DFL);
       (void)signal(SIGCHLD, SIG_DFL);
+      (void)signal(SIGTTIN, SIG_DFL);
 
-      while (dup2(2, 1) < 0 && (errno == EINTR))
-	;
+      TEMP_FAILURE_RETRY(dup2(2, 1));
       closeall();
       if (run_mode)
 	{
@@ -548,8 +616,7 @@ int run_single(const char *prg, const char *arg0, int spl)
       _exit(1);
     }
 
-   while ((waitpid(pid, &r, 0) == (pid_t)-1) && (errno == EINTR))
-     ;
+   TEMP_FAILURE_RETRY(waitpid(pid, &r, 0));
    callsplash(spl, prg, arg);
    return WIFEXITED(r) ? WEXITSTATUS(r) : (WIFSIGNALED(r) ? 1 : 255);
 }
@@ -557,17 +624,27 @@ int run_single(const char *prg, const char *arg0, int spl)
 void do_forward(void)
 {
   char buf[4096], *b;
-  int r, rr;
+  ssize_t r, rr;
   setsid();
-  while ((r = read(0, buf, sizeof(buf))) > 0)
+  while ((r = read(0, buf, sizeof(buf))))
     {
+      if (r < 0)
+	{
+	  if (errno == EINTR)
+	    continue;
+	  perror("\n\rstartpar: forward read");
+	  break;
+	}
       b = buf;
       while (r > 0)
 	{
 	  rr = write(1, b, r);
 	  if (rr == -1)
 	    {
-	      perror("write");
+	      if (errno == EINTR)
+		continue;
+	      perror("\n\rstartpar: forward write");
+	      break;
 	      rr = r;
 	    }
 	  r -= rr;
@@ -623,8 +700,9 @@ void flushbuf(void)
 
 void detach(struct prg *p, const int store)
 {
-  int r;
   int flags = fcntl(p->fd, F_GETFL);
+  ssize_t r;
+  pid_t pid;
 
   if (flags > 0)
     flags |= FNONBLOCK;
@@ -632,8 +710,14 @@ void detach(struct prg *p, const int store)
     flags = FNONBLOCK;
 
   fcntl(p->fd, F_SETFL, flags);
-  while ((r = read(p->fd, p->buf, sizeof(p->buf))) > 0)
+  while ((r = read(p->fd, p->buf, sizeof(p->buf))))
     {
+      if (r < 0)
+	{
+	  if (errno == EINTR)
+	    continue;
+	  break;
+	}
       p->len = r;
       if (store)
 	storebuf(p);
@@ -644,30 +728,66 @@ void detach(struct prg *p, const int store)
   fcntl(p->fd, F_SETFL, flags);
   if (r == -1 && errno == EWOULDBLOCK)
     {
-      if ((r = fork()) == 0)
+      if ((pid = fork()) == 0)
 	{
-	  while (dup2(p->fd, 0) < 0 && (errno == EINTR))
-	    ;
-	  while (dup2(2, 1) < 0 && (errno == EINTR))
-	    ;
+	  TEMP_FAILURE_RETRY(dup2(p->fd, 0));
+	  TEMP_FAILURE_RETRY(dup2(2, 1));
 	  closeall();
 	  execlp(myname, myname, "-f", "--", p->name, NULL);
 	  do_forward();
 	}
-      if (r == -1)
+      if (pid == -1)
 	perror("fork");
     }
   close(p->fd);
   p->fd = 0;
 }
 
-static void sigchld(int sig __attribute__ ((unused)))
+static struct prg *interactive_task;
+static volatile int active;
+static void sigchld(int sig attribute((unused)))
 {
-  char c = 0;
-  write(pidpipe[1], &c, 1);
+  const int old_errno = errno;
+  int status;
+  pid_t pid;
+ 
+  while ((pid = waitpid(-1, &status, WNOHANG|WUNTRACED)) != 0)
+    {
+      int s;
+
+      if (errno == ECHILD)
+	break;
+
+      if (pid < 0)
+	continue;
+
+      if (pid == splashpid)
+	{
+	  splashpid = (pid_t)0;
+	  continue;
+	}
+
+      for (s = 0; s < par; s++)
+	{
+	  struct prg * p = prgs + s;
+	  if (p == interactive_task)
+	    continue;
+	  if (p->pid != pid)
+	    continue;
+	  if (p->flags & PRUNNING)
+	    {
+	      p->status = status;
+	      p->flags |= PFINISHED;
+	      break;
+	    }
+	}
+    }
+
+  signaled = 1;
+  errno = old_errno;
 }
 
-static void sigwinch(int sig __attribute__ ((unused)))
+static void sigwinch(int sig attribute((unused)))
 {
   if (ioctl(0, TIOCGWINSZ, &wz) < 0)
     {
@@ -678,6 +798,7 @@ static void sigwinch(int sig __attribute__ ((unused)))
   if (wz.ws_col == 0) wz.ws_col = 80;
   snprintf(sz.env_row, sizeof(sz.env_row), "LINES=%d",   wz.ws_row);
   snprintf(sz.env_col, sizeof(sz.env_col), "COLUMNS=%d", wz.ws_col);
+  kill(-1, SIGWINCH);
 }
 
 void usage(int status)
@@ -704,15 +825,17 @@ int main(int argc, char **argv)
   int c, i, num;
   int limit;
   int *resvec;
-  fd_set rset;
   struct timeval tv;
   struct prg *p;
-  char pipebuf[16];
   struct prg *gtimo_running = 0;
-  struct prg *interactive_task = NULL;
   char *prev_level = getenv("PREVLEVEL");
   char *run_level = getenv("RUNLEVEL");
   char *splashopt = 0;
+  sigset_t nmask, omask, smask;
+
+  (void)sigemptyset(&nmask);
+  (void)sigaddset(&nmask, SIGHUP);
+  sigprocmask(SIG_UNBLOCK, &nmask, NULL);
 
   (void)signal(SIGCHLD, SIG_DFL);
   numcpu = sysconf(_SC_NPROCESSORS_ONLN);
@@ -819,7 +942,7 @@ int main(int argc, char **argv)
 	exit(0);
 
       if (par == 0)
-	par = 4;
+	par = 8;
       if (par > argc)				/* not more than the number of all scripts */
 	par = argc;
 
@@ -893,8 +1016,10 @@ int main(int argc, char **argv)
   if (tcgetattr(0, &tio))
     {
       perror("tcgetattr");
-      exit(1);
+      tcgetattr(2, &tio);
     }
+  cfmakeraw(&tio);
+
   if (ioctl(0, TIOCGWINSZ, &wz) == 0)
     wzok = 1;
   if (wz.ws_row == 0) wz.ws_row = 24;
@@ -905,16 +1030,13 @@ int main(int argc, char **argv)
   snprintf(sz.env_row, sizeof(sz.env_row), "LINES=%d",   wz.ws_row);
   snprintf(sz.env_col, sizeof(sz.env_col), "COLUMNS=%d", wz.ws_col);
 
-  if (pipe(pidpipe))
-    {
-      perror("pipe");
-      exit(1);
-    }
-  fcntl(pidpipe[0], F_SETFL, FNONBLOCK);
-  fcntl(pidpipe[1], F_SETFL, FNONBLOCK);
   sa.sa_handler = sigchld;
   sa.sa_flags = SA_RESTART;
   (void)sigemptyset(&sa.sa_mask);
+  (void)sigaddset(&sa.sa_mask, SIGCHLD);
+  (void)sigaddset(&sa.sa_mask, SIGWINCH);
+  (void)sigaddset(&sa.sa_mask, SIGUSR1);
+  (void)sigaddset(&sa.sa_mask, SIGUSR2);
   if (sigaction(SIGCHLD, &sa, 0))
     {
       perror("sigchld sigaction");
@@ -941,22 +1063,40 @@ int main(int argc, char **argv)
     }
 #endif
 
+  /* lock us into memory */
+  if (geteuid() == 0)
+    mlockall(MCL_CURRENT|MCL_FUTURE);
+  errno = 0;
+
   gettimeofday(&glastio, 0);
   limit = checklimit(inpar, isstart);
   lastlim = glastio;
+
+  (void)sigemptyset(&nmask);
+  (void)sigaddset(&nmask, SIGCHLD);
+  sigprocmask(SIG_BLOCK, &nmask, &omask);
+
+  (void)sigfillset(&smask);
+  (void)sigdelset(&smask, SIGHUP);
+  (void)sigdelset(&smask, SIGINT);
+  (void)sigdelset(&smask, SIGQUIT);
+  (void)sigdelset(&smask, SIGTERM);
+  (void)sigdelset(&smask, SIGCHLD);
+
   for (;;)
     {
-      int active = 0;
-      int maxfd = -1;
-      int last = -1;
-      pid_t pid = 0;
-      int r = 0, s;
-      long diff;
 #ifdef CHECK_FORDEVPTS
       int devpts = 0;
 #endif
+      int maxfd = -1;
+      int last = -1;
+      int r, s;
+      long diff;
+      pid_t pid;
+      fd_set rset;
 
       gettimeofday(&now, 0);
+      pid = (pid_t)-1;
       FD_ZERO(&rset);
       tv = now;
 
@@ -975,22 +1115,16 @@ int main(int argc, char **argv)
 #endif
       for (s = 0; s < par; s++)			/* never leave this with break!! */
 	{
-	account:				/* for the new process below */
-#ifdef CHECK_FORDEVPTS
-	  if (!devpts)
-	    devpts = checkdevpts();
-#endif
 	  p = prgs + s;
 	  if (p == interactive_task)
 	    continue;				/* don't count this here */
-	  if (p->fd || p->pid)
-	    active++;				/* count all running procs */
 	  if (p->fd == 0)
 	    {
 	      if (interactive_task)
 		continue;			/* dont't start new processes */
 	      if (num >= argc)
 		continue;			/* nothing to do */
+	    redo:
 	      if (p->pid == 0)
 		{
 		  if (active >= limit)
@@ -1010,31 +1144,40 @@ int main(int argc, char **argv)
 		  }
 		  p->splashadd = calcsplash(num, argc, splashopt);
 		  p->num = num++;
-		  if (interactive_task)
-		    continue;			/* don't start this here */
 #ifdef CHECK_FORDEVPTS
 		  if (!devpts)
-		    {
-		      interactive_task = p;	/* no /dev/pts, treat as interactive */
-		      continue;
-		    }
+		    interactive_task = p;	/* no /dev/pts, treat as interactive */
 #endif
+		  if (interactive_task)
+		    continue;			/* don't start this here */
 		  run(p);
 		  if (p->pid == 0)
 		    {
 		      resvec[p->num] = 1;
 		      if (run_mode)
 			finish_task(nodevec[p->num]);
+		      p->status = 0;
+		      p->flags = 0;
+		      if (p->fd)
+			{
+			  close(p->fd);
+			  p->fd = 0;
+			}
+		      goto redo;		/* this one is free again */
+		    }
+		  else
+		    {
+		      active++;			/* count this active job */
 		    }
 		  gettimeofday(&now, 0);
 		  tv = now;
-		  goto account;			/* take the new process into account */
 		}
-	      continue;
 	    }
+
 	  FD_SET(p->fd, &rset);
 	  if (p->fd > maxfd)
 	    maxfd = p->fd;
+
 	  if (p->len == 0)
 	    continue;
           if ((last < 0) || timercmp(&tv,&p->lastio,>))
@@ -1042,6 +1185,7 @@ int main(int argc, char **argv)
 	      last = s;
 	      tv = p->lastio;
 	    }
+
 	} /* for (s = 0; s < par; s++) */
 
       if (interactive_task)
@@ -1049,9 +1193,11 @@ int main(int argc, char **argv)
 	  if (active == 0)
 	    {
 	      p = interactive_task;
+	      p->flags |= PRUNNING;
 	      resvec[p->num] = run_single(p->name, p->arg0, p->splashadd);
 	      if (run_mode)
 		finish_task(nodevec[p->num]);
+	      p->flags = 0;
 	      p->pid = 0;
 	      p->fd = 0;
 	      interactive_task = NULL;
@@ -1063,14 +1209,10 @@ int main(int argc, char **argv)
 	{
 	  if (num < argc)
 	    fprintf(stderr, "ERROR: not all processed (%d of %d)\n", num, argc);
-#if DEBUG
-	  if ((pid = waitpid(-1, &r, (maxfd < 0 ? 0 : WNOHANG)|WUNTRACED)) > 0)
-	    fprintf(stderr, "ERROR: not all processes are checked\n");
-#endif
 	  break;
 	}
 #if DEBUG
-      fprintf(stderr, "active = %d\n", active);
+      fprintf(stderr, "active = %d, maxfd = %d, last = %d\n", active, maxfd, last);
 #endif
       if (active == 1 && last >= 0)
 	{
@@ -1080,68 +1222,81 @@ int main(int argc, char **argv)
 	      writebuf(p);
 	      continue;
 	    }
+	  else
+	    {
+	      p->status = r;
+	      p->flags |= PFINISHED;
+	      signaled = 1;
+	    }
+	}
+      else
+	{
+	  const struct timespec zero = {0,0};
+	  const int old_errno = errno;
+	  (void)pselect(0,0,0,0,&zero,&smask);
+	  errno = old_errno;
 	}
 
-      FD_SET(pidpipe[0], &rset);
-      /* drain the pidpipe */
-      while ((c = read(pidpipe[0], pipebuf, sizeof pipebuf)) > 0)
-	;
-
-      if (pid == 0)
-	pid = waitpid(-1, &r, (maxfd < 0 ? 0 : WNOHANG)|WUNTRACED);
-
-      if (pid > 0)
+      if (signaled)
 	{
-	  if (pid == splashpid)
-	    splashpid = (pid_t)0;
+	  signaled = 0;
+
 	  for (s = 0; s < par; s++)
 	    {
 	      p = prgs + s;
-	      if (p->pid == pid)
-		{
-		  if (WIFSTOPPED(r))
-		    {
-		      if (WSTOPSIG(r) == SIGTTIN)
-			{
-			  pid_t pg = getpgid(pid);
-			  if (pg > 0)
-			    killpg(pg, SIGCONT);
-			}
-		      continue;
-		    }
-		  callsplash(p->splashadd, p->name, arg);
-		  resvec[p->num] = WIFEXITED(r) ? WEXITSTATUS(r) : (WIFSIGNALED(r) ? 1 : 255);
-		  if (run_mode)
-		    finish_task(nodevec[p->num]);
-		  p->pid = 0;
-		  if (gtimo_running == p)
-		    {
-		      writebuf(p);
-		      if (p->fd)
-			detach(p, GTIMO_OFFL);
-		      flushbuf();
-		      gtimo_running = 0;
-		    }
-		  else if (gtimo_running)
-		    {
-		      storebuf(p);
-		      if (p->fd)
-			detach(p, GTIMO_USED);
-		    }
-		  else
-		    {
-		      writebuf(p);
-		      if (p->fd)
-			detach(p, GTIMO_OFFL);
-		    }
-		  break;
-		}
-	    } /* for (s = 0; s < par; s++) */
-	  continue;
-	}
 
-      if (maxfd < 0)
-	continue;				/* start new processes? */
+	      if ((p->flags & PFINISHED) == 0)
+		continue;
+	      r = p->status;
+
+	      if (WIFSTOPPED(r))
+		{
+		  pid_t pg = getpgid(p->pid);
+		  if (pg > 0)
+		    killpg(pg, SIGCONT);
+		  else
+		    kill(-p->pid, SIGCONT);
+		  p->flags &= ~PFINISHED;
+		  continue;
+		}
+
+	      callsplash(p->splashadd, p->name, arg);
+	      resvec[p->num] = WIFEXITED(r) ? WEXITSTATUS(r) : (WIFSIGNALED(r) ? 1 : 255);
+	      if (run_mode)
+		finish_task(nodevec[p->num]);
+	      p->status = 0;
+	      p->flags = 0;
+	      p->pid = 0;
+	      active--;
+
+	      if (gtimo_running == p)
+		{
+		  writebuf(p);
+		  if (p->fd) detach(p, GTIMO_OFFL);
+		  flushbuf();
+		  gtimo_running = 0;
+		}
+	      else if (gtimo_running)
+		{
+		  storebuf(p);
+		  if (p->fd) detach(p, GTIMO_USED);
+		}
+	      else
+		{
+		  writebuf(p);
+		  if (p->fd) detach(p, GTIMO_OFFL);
+		}
+#ifdef START_OUTRIGHT
+	      break;		/* one free slot ... start outright new process */
+#endif
+	    } /* for (s = 0; s < par; s++) */
+
+#ifdef CHECK_FORDEVPTS
+	  if (!devpts)
+	    devpts = checkdevpts();
+#endif
+	  continue;		/* start new processes */
+	}
 
       if (timo >= 0)
         tv.tv_sec += timo;
@@ -1162,8 +1317,8 @@ int main(int argc, char **argv)
       r = 0;
       if (timo >= 0 || isgtimo)
 	{
-	  int setfd = (pidpipe[0] > maxfd) ? pidpipe[0] : maxfd;
 	  struct timeval wait;
+	  struct timespec spec;
 
 	  timersub(&tv, &now, &wait);
 	  if (wait.tv_usec < 0)
@@ -1183,7 +1338,8 @@ int main(int argc, char **argv)
 #if DEBUG
 	      fprintf(stderr, "going into select1 %d %ld %ld\n", last, wait.tv_sec, wait.tv_usec);
 #endif
-	      r = select(setfd + 1, &rset, 0, 0, (last >= 0 || check) ? &wait : 0);
+	      TIMEVAL_TO_TIMESPEC(&wait, &spec);
+	      r = pselect(maxfd + 1, &rset, 0, 0, (last >= 0 || check) ? &spec : 0, &smask);
 
 	      if (check && (r == 0))		/* ... but do not throw out messages to early!!! */
 		continue;
@@ -1192,19 +1348,19 @@ int main(int argc, char **argv)
 	    {
 	      wait.tv_sec  = 0;			/* Avoid looping around (does this ever happen?) */
 	      wait.tv_usec = 20*1000;
-	      r = select(setfd + 1, &rset, 0, 0, last >= 0 ? &wait : 0);
+	      TIMEVAL_TO_TIMESPEC(&wait, &spec);
+	      r = pselect(maxfd + 1, &rset, 0, 0, last >= 0 ? &spec : 0, &smask);
 	    }
 	}
       else
 	{
-	  int setfd = (pidpipe[0] > maxfd) ? pidpipe[0] : maxfd;
-	  r = select(setfd + 1, &rset, 0, 0, 0);
+	  r = pselect(maxfd + 1, &rset, 0, 0, 0, &smask);
 	}
 
       if (r == -1)
 	{
 	  if (errno == EINTR)
-	    continue;
+	    continue;		/* SIGCHLD has happen in most cases */
 	  perror("select");
 	  exit(1);
 	}
@@ -1226,14 +1382,20 @@ int main(int argc, char **argv)
 		continue;
 	      if (!FD_ISSET(p->fd, &rset))
 		continue;
-	      r = read(p->fd, p->buf + p->len, sizeof(p->buf) - p->len);
+	      r = (int)TEMP_FAILURE_RETRY(read(p->fd, p->buf + p->len, sizeof(p->buf) - p->len));
 	      if (r <= 0)
 		{
 		  if (!gtimo_running || p == gtimo_running)
 		    writebuf(p);
-		  close(p->fd);
-		  p->fd = 0;
-		  break;
+		  if (r < 0)
+		    {
+		      close(p->fd);
+		      p->fd = 0;
+#ifdef START_OUTRIGHT
+		      break;		/* one free slot ... start outright new process */
+#endif
+		    }
+		  continue;
 		}
 	      p->len += r;
 	      if (p->len == sizeof(p->buf))
@@ -1253,11 +1415,15 @@ int main(int argc, char **argv)
 		  p->len = sizeof(p->buf) - p->len;
 		}
 	      p->lastio = now;
+
 	    } /* for (s = 0; s < par; s++) */
 	}
+
     } /* for (;;) */
 
- finished:
+  sigprocmask(SIG_SETMASK, &omask, NULL);
+
+finished:
   waitsplash();
   if (run_mode)
     print_run_result(resvec, nodevec, run_mode);
